unit VPUtils;

interface

{$define BEST}

Uses
    Controls;

const
     LIMIT	       = 2202;
     SOFTLIMIT     = 2000;
     NUM_DEFAULTS  = 192;
     NUM_LISTITEMS = 34;
     INVALID       = -987;
     KEY001        = 68150631;
     KEY002        = 9371;

type
    TSchedule = record
        SDate: TDate;     // Start date
        STime: TTime;     // Start time
        ETime: TTime;     // End time
        Durn: integer;    // Duration in minutes
        Channel: integer; // Channel
    end;

    Ts_tab = record    // Start time & duration table record
	   start: integer;
	   leng: integer;
    end;
    Ps_tab = ^Ts_tab; // Pointer to ....
    Tar_s_tab = Array[0 .. LIMIT-1] of Ts_tab;
    Tar_s_tab_dflts = Array[0 .. NUM_DEFAULTS-1] of Ts_tab;
    Tar_list = Array[0 .. NUM_LISTITEMS-1] of integer;

// Main Function prototypes
function Decode(const VPCode: LongInt; const RefDate: TDate; var Sched: TSchedule): LongInt;
function Encode(const RefDate: TDate; var Sched: TSchedule): LongInt;
// Encoding & Decoding function prototypes
function F1(inval, key: LongInt): LongInt;
function NumDigits(const inval: LongInt; var pwr: LongInt): Integer;
function Mixup(x, y: LongInt): LongInt;
procedure Offset(const day, year: integer; const top: LongInt;
                 var ofout: integer; var topout: LongInt);
function EncodeBot3(const day, rem: integer): integer;
procedure DeInterleave(const t8c5: LongInt; const t2c1: Integer; var tval: LongInt; var cval: integer);
procedure Interleave(const tblidx: LongInt; const cval: integer; var t8c5: LongInt; var t2c1: integer);
function EndTime(start, dur: integer): integer;
procedure FourDigitYear(var ADate: TDate);
function BestTimes(var rstart, rduration: integer; var heuristic: Longword): LongInt;
function DiffMinutes(startT, endT: integer): integer;
function ScanFor(const s, d: integer ): LongInt;
function MapTop(const day, year: integer; const top: LongInt; const digits: integer): LongInt;
function UnmapTop(const day, year: integer; const top: LongInt; const digits: integer): LongInt;
function AddTime(const start, dur: integer): Integer;
procedure Lookup(const i: LongInt; var outtime, outdur: integer);
procedure SetTables;
procedure Fill(const index, time, dur: integer);
procedure FillQHB(const index, dur: integer);
procedure FillHHB(const index, dur: integer);

implementation

Uses
    SysUtils;

var
   initializer: integer;
   change: integer;       // probably could be a boolean
   slottable: Tar_s_tab;  // populated at runtime - PP

const
  // insert movie (and other) structured tables
  tlist: Tar_list = (5, 10,20, 25,
		             0, 5, 10, 15, 20, 25,	// -5
                     0, 5, 10, 15, 20, 25,	// -10
		             0, 5, 10, 15, 20, 25,	// -15
		             0, 5, 10, 15, 20, 25,	// -20
		             0, 5, 10, 15, 20, 25);	// -25

  dlist: Tar_list = (0,    0,   0,   0,
		            -5,   -5,  -5,  -5,  -5,  -5,
		            -10, -10, -10, -10, -10, -10,
		            -15, -15, -15, -15, -15, -15,
		            -20, -20, -20, -20, -20, -20,
		            -25, -25, -25, -25, -25, -25);

  // first 192 entries are erratic, so use pre-loaded defaults
  dflt_slottable: Tar_s_tab_dflts =
    ((start:1830;leng:30),
    (start:1600;leng:30),
    (start:1930;leng:30),
    (start:1630;leng:30),
    (start:1530;leng:30),
    (start:1730;leng:30),
    (start:1800;leng:30),
    (start:1430;leng:30),
    (start:1900;leng:30),
    (start:1700;leng:60),
    (start:1400;leng:30),
    (start:2030;leng:30),
    (start:1700;leng:30),
    (start:1600;leng:120),
    (start:2000;leng:30),
    (start:1500;leng:30),
    (start:2000;leng:120),
    (start:2100;leng:120),
    (start:2000;leng:60),
    (start:1800;leng:120),
    (start:1900;leng:60),
    (start:2200;leng:60),
    (start:2100;leng:60),
    (start:1400;leng:120),
    (start:1500;leng:60),
    (start:2200;leng:120),
    (start:1130;leng:30),
    (start:1100;leng:30),
    (start:2300;leng:30),
    (start:1600;leng:60),
    (start:2100;leng:90),
    (start:2100;leng:30),
    (start:1230;leng:30),
    (start:1330;leng:30),
    (start:930;leng:30),
    (start:1300;leng:60),
    (start:2130;leng:30),
    (start:1200;leng:60),
    (start:1000;leng:120),
    (start:1800;leng:60),
    (start:2200;leng:30),
    (start:1200;leng:30),
    (start:800;leng:30),
    (start:830;leng:30),
    (start:1700;leng:120),
    (start:900;leng:30),
    (start:2230;leng:30),
    (start:1030;leng:30),
    (start:1900;leng:120),
    (start:730;leng:30),
    (start:2300;leng:60),
    (start:1000;leng:60),
    (start:700;leng:30),
    (start:1300;leng:30),
    (start:700;leng:120),
    (start:1100;leng:60),
    (start:1400;leng:60),
    (start:1000;leng:30),
    (start:800;leng:120),
    (start:2330;leng:30),
    (start:1300;leng:120),
    (start:1200;leng:120),
    (start:900;leng:120),
    (start:630;leng:30),
    (start:1800;leng:90),
    (start:600;leng:30),
    (start:530;leng:30),
    (start:0;leng:30),
    (start:2330;leng:120),
    (start:2200;leng:90),
    (start:1300;leng:90),
    (start:900;leng:60),
    (start:1630;leng:90),
    (start:1600;leng:90),
    (start:1430;leng:90),
    (start:2000;leng:90),
    (start:1830;leng:90),
    (start:600;leng:60),
    (start:1200;leng:90),
    (start:30;leng:30),
    (start:130;leng:120),
    (start:0;leng:60),
    (start:1700;leng:90),
    (start:0;leng:120),
    (start:800;leng:60),
    (start:700;leng:60),
    (start:2130;leng:120),
    (start:500;leng:30),
    (start:1530;leng:90),
    (start:1130;leng:120),
    (start:1100;leng:120),
    (start:830;leng:90),
    (start:2230;leng:90),
    (start:900;leng:90),
    (start:2130;leng:90),
    (start:1630;leng:120),
    (start:2330;leng:60),
    (start:100;leng:120),
    (start:1400;leng:90),
    (start:130;leng:30),
    (start:330;leng:120),
    (start:1500;leng:90),
    (start:1500;leng:120),
    (start:2300;leng:120),
    (start:1900;leng:90),
    (start:800;leng:90),
    (start:430;leng:30),
    (start:300;leng:30),
    (start:1330;leng:120),
    (start:1000;leng:90),
    (start:700;leng:90),
    (start:100;leng:30),
    (start:2330;leng:90),
    (start:330;leng:30),
    (start:200;leng:30),
    (start:2230;leng:120),
    (start:400;leng:30),
    (start:600;leng:120),
    (start:400;leng:120),
    (start:230;leng:30),
    (start:630;leng:90),
    (start:30;leng:60),
    (start:2230;leng:60),
    (start:100;leng:60),
    (start:30;leng:120),
    (start:2300;leng:90),
    (start:1630;leng:60),
    (start:830;leng:60),
    (start:0;leng:90),
    (start:1930;leng:120),
    (start:930;leng:120),
    (start:2030;leng:90),
    (start:500;leng:60),
    (start:1730;leng:60),
    (start:200;leng:120),
    (start:1930;leng:90),
    (start:930;leng:90),
    (start:1730;leng:120),
    (start:630;leng:120),
    (start:1830;leng:60),
    (start:1430;leng:60),
    (start:1130;leng:90),
    (start:30;leng:90),
    (start:830;leng:120),
    (start:1030;leng:90),
    (start:1430;leng:120),
    (start:100;leng:90),
    (start:730;leng:120),
    (start:2030;leng:120),
    (start:300;leng:90),
    (start:300;leng:120),
    (start:1330;leng:90),
    (start:1230;leng:90),
    (start:230;leng:90),
    (start:2130;leng:60),
    (start:1130;leng:60),
    (start:1830;leng:120),
    (start:630;leng:60),
    (start:530;leng:60),
    (start:200;leng:60),
    (start:1530;leng:120),
    (start:730;leng:60),
    (start:600;leng:90),
    (start:1730;leng:90),
    (start:400;leng:60),
    (start:730;leng:90),
    (start:430;leng:90),
    (start:430;leng:60),
    (start:130;leng:90),
    (start:1230;leng:120),
    (start:130;leng:60),
    (start:230;leng:120),
    (start:1930;leng:60),
    (start:300;leng:60),
    (start:1030;leng:120),
    (start:200;leng:90),
    (start:330;leng:60),
    (start:500;leng:120),
    (start:930;leng:60),
    (start:230;leng:60),
    (start:2030;leng:60),
    (start:400;leng:90),
    (start:1530;leng:60),
    (start:430;leng:120),
    (start:1330;leng:60),
    (start:1230;leng:60),
    (start:330;leng:90),
    (start:1030;leng:60),
    (start:500;leng:90),
    (start:530;leng:120),
    (start:530;leng:90),
    (start:1100;leng:90));

// ------------------------------------------------------------------- /
// F1 : First Function                                                 /
//      Used in decoding.                                              /
// ------------------------------------------------------------------- /
function F1(inval, key: LongInt): LongInt;
var
   hashval, pwr: LongInt;
   ndigits: integer;
begin
   ndigits := NumDigits(inval,pwr);
   if ndigits > 8 then
      raise Exception.Create(format('%d has %d digits !',[inval,ndigits]));
   pwr := pwr div 10;
   hashval := inval;
   repeat
        hashval := mixup(hashval,key) mod (pwr * 10);
   until not (hashval < pwr);
   Result := hashval;
end;

// ------------------------------------------------------------------- /
// NumDigits : How many digits in a number?                            /
//      Used in decoding.                                              /
// ------------------------------------------------------------------- /
function NumDigits(const inval: LongInt; var pwr: LongInt): Integer;
begin
   Result := 0;
   pwr := 1;
   while (inval >= pwr) do begin
     Result := Result + 1;
     pwr := pwr * 10;
   end;
end;

// ------------------------------------------------------------------- /
// Mixup : Function that performs initial scrambling                   /
//         Used in decoding.                                           /
// ------------------------------------------------------------------- /
function Mixup(x, y: LongInt): LongInt;
var
    i, j, k: LongInt;
    a: array[0..11] of LongInt;
    b: array[0..11] of LongInt;
    output: array[0..19] of LongInt;
begin
    Result := 0;
    // get the digits of x into a[]
    j := x;
    for i := 0 to 8 do begin
        k := j mod 10;
        a[i] := k;
        j := (j-k) div 10;
    end;

    // get the digits of y into b[]
    j := y ;
    for i := 0 to 8 do begin
        k := j mod 10;
        b[i] := k;
        j := (j-k) div 10;
        output[i] := 0;
    end;

    for i := 0 to 8 do begin
        for j := 0 to 8 do
           output[i+j] := output[i+j] + (b[j] * a[i]);
    end;

    j := 1;
    for i := 0 to 8 do begin
        Result := Result + (j * (output[i] mod 10));
        j := j * 10 ;
    end;
end;

// ------------------------------------------------------------------- /
// Offset : ???????                                                    /
//          Used in decoding.                                          /
// ------------------------------------------------------------------- /
procedure Offset(const day, year: integer; const top: LongInt;
                 var ofout: integer; var topout: LongInt);
var
    i, digits, off: integer;
    t, tx, pwr: LongInt;
begin
    pwr := 1;
    digits := 0;
    while top >= pwr do begin
       inc(digits);
       pwr := pwr * 10;
    end;
    pwr := pwr div 10 ;

    t := top;
    tx := top;     // t = tx = top;
    off := 0;
    while tx > 0 do begin
      off := off + integer(tx mod 10);
      tx := tx div 10;
    end;

    repeat
       for i := 0 to (year mod 16) do    // for(i=0; i<=(year%16); i++)
         off := off + integer((MapTop(day,i,t,digits) mod 10));
       t := MapTop(day,year,t,digits) ;
    until not (t < pwr);

    ofout := off mod 32;
    topout := t;
end;

// ------------------------------------------------------------------- /
// EncodeBot3 : ???????                                                /
//              Used in encoding.                                      /
// ------------------------------------------------------------------- /
function EncodeBot3(const day, rem: integer): integer;
begin
     Result := rem + (32 * (day - 1)) + 1;
end;

// ------------------------------------------------------------------- /
// MapTop : ???????                                                    /
//          Used in decoding & encoding.                               /
// ------------------------------------------------------------------- /
function MapTop(const day, year: integer; const top: LongInt; const digits: integer): LongInt;
var
    d2, d1, d0, y: LongInt ;
    n2, n1, n0, f3, f2, f1, f0: LongInt;
begin
    y := year mod 16 ;
    d2 := top div 100 ;
    d1 := (top mod 100) div 10 ;
    d0 := top mod 10 ;

    f0 := 1;
    f1 := (y + 1) mod 10;
    f2 := ( ((y+1)*(y+2)) div 2 ) mod 10 ;
    f3 := ( ((y+1)*(y+2)*(y+3)) div 6 ) mod 10 ;

    n0 := 0; n1 := 0; n2 := 0;
    case Digits of
         1 : begin
                 n0 := ( (d0*f0) + (day*f1) ) mod 10 ;
                 n1 := 0;
                 n2 := 0;
             end;

         2 : begin
                  n0 := ( (d0*f0) + (d1*f1) + (day*f2) ) mod 10 ;
                  n1 := ( (d1*f0) + (day*f1) ) mod 10 ;
                  n2 := 0;
             end;
         3 : begin
                  n0 := ( (d0*f0) + (d1*f1) + (d2*f2) + (day*f3) ) mod 10 ;
                  n1 := ( (d1*f0) + (d2*f1) + (day*f2) ) mod 10 ;
                  n2 := ( (d2*f0) + (day*f1) ) mod 10 ;
             end;
    end;
    Result :=  (100*n2) + (10*n1) + n0 ;
end;

// ------------------------------------------------------------------- /
// MapTop : ???????                                                    /
//          Used in encoding.                                          /
// ------------------------------------------------------------------- /
function UnmapTop(const day, year: integer; const top: LongInt; const digits: integer): LongInt;
var
    d2, d1, d0, y: LongInt ;
    n2, n1, n0, f3, f2, f1, f0, p3, p2, p1: LongInt;
begin
	d2 := top div 100 ;
	d1 := (top mod 100) div 10 ;
	d0 := top mod 10 ;
	// generate key (P3P2P1F0) and reverse key (F3F2F1F0)
	f0 := 1;
	y := year mod 16 ;
	p1 := (y + 1) mod 10;
	f1 := 10 - p1;
	p2 := ( ((y+1)*(y+2)) div 2 ) mod 10 ;
	f2 := 10 - (p2 + f1*p1) mod 10;
	p3 := ( ((y+1)*(y+2)*(y+3)) div 6 ) mod 10 ;
	f3 := 10 - (p3 + f1*p2 + f2*p1) mod 10;
    n0 := 0; n1 := 0; n2 := 0;
	case digits of
         1 : begin
                n0 := ( (d0*f0) + (day*f1) ) mod 10 ;
                n1 := 0;
                n2 := 0;
             end;
         2 : begin
                n0 := ( (d0*f0) + (d1*f1) + (day*f2) ) mod 10 ;
                n1 := ( (d1*f0) + (day*f1) ) mod 10 ;
                n2 := 0;
             end;
         3 : begin
                n0 := ( (d0*f0) + (d1*f1) + (d2*f2) + (day*f3) ) mod 10 ;
		        n1 := ( (d1*f0) + (d2*f1) + (day*f2) ) mod 10 ;
		        n2 := ( (d2*f0) + (day*f1) ) mod 10 ;
             end;
    end;
	Result :=  (100*n2) + (10*n1) + n0 ;
end;

// ---------------------------------------------------------------------------- /
// Interleave and DeInterleave routines.                                       /
//                                                                              /
// The American routine shuffles the bits differently from the British.         /
//                     |                                                        /
// US:      T8|C5|T7|C4|T6|T5|T4|C3|C2|T3      T2|C1|T1|C0|T0                   /
// UK:     T10|T9|T8|T7|T6|T5|T4|C3|C2|T3      T2|C1|T1|C0|T0                   /
//                     |                                                        /
//                                                                              /
// Break points between 4,5 and 6 and 7 digit codes are where the left inter-   /
// leave gets to 9,99,999 or 9999. Assume the interleave of T2,T1,T0 and C1,C0  /
// is all-ones.                                                                 /
//                                                                              /
// In the UK:                                                                   /
//         T10|T9|T8|T7|T6|T5|T4|C3|C2|T3                                       /
// 9(0x9)    0| 0| 0| 0| 0| 0| 1| 0| 0| 1  -> tblidx=00000011111, cval=000011.  /
// 99(0x63)  0| 0| 0| 1| 1| 0| 0| 0| 1| 1  -> tblidx=00011001111, cval=000111.  /
// 999(0x3E7)1| 1| 1| 1| 1| 0| 0| 1| 1| 1  -> tblidx=11111001111, cval=001111.  /
//                                                                              /
// The 4-digit break is at tblidx=31 on channel 4, or if we trade T4 to allow   /
// C3 and C2 to be set, then it is at tblidx=15 on channel 16.                  /
// The 5-digit break is at tblidx=207 on channel 8, or if we trade T4 to allow  /
// C3 to be set, then it is at tblidx=191 on channel 16.                        /
// The 6-digit break is at tblidx=1999 on channel 16, so no trade can be done.  /
//                                                                              /
// In the US:                                                                   /
//          T8|C5|T7|C4|T6|T5|T4|C3|C2|T3                                       /
// 9(0x9)    0| 0| 0| 0| 0| 0| 1| 0| 0| 1  -> tblidx=00000011111, cval=000011.  /
// 99(0x63)  0| 0| 0| 1| 1| 0| 0| 0| 1| 1  -> tblidx=00001001111, cval=010111.  /
// 999(0x3E7)1| 1| 1| 1| 1| 0| 0| 1| 1| 1  -> tblidx=00111001111, cval=111111.  /
//                                                                              /
// The 4-digit break is at tblidx=31 on channel 4, or if we trade T4 to allow   /
// C3 and C2 to be set, then it is at tblidx=15 on channel 16.                  /
// The 5-digit break is at tblidx=79 on channel 24, or if we trade T6 to allow  /
// C3 (and T5 and T4) to be set, then it is at tblidx=63 on channel 32. Or if   /
// we trade C4 to allow T5 and T4 and C3 to be set, then it is at tblidx=127 on /
// channel 16.                                                                  /
// The 6-digit break is at tblidx=463 on channel 64, or if we trade C4 to allow /
// T5 and T4 to be set, then it is at tblidx=511 on channel 48.                 /
//                                                                              /
//                                                                              /
// Notice that if the programme's code doesn't use bits C4 and up (i.e          /
// channels 1-16), and if the timeslot index number doesn't use bits T7 and     /
// up (tblidx <= 127) then US VCR+ and UK videoplus will produce the same       /
// codes.                                                                       /
// ---------------------------------------------------------------------------- /

// ------------------------------------------------------------------- /
// DeInterleave : Used in decoding                                     /
// ------------------------------------------------------------------- /
procedure DeInterleave(const t8c5: LongInt; const t2c1: Integer; var tval: LongInt; var cval: integer);
var
    cc : LongInt;
    tt, b: LongInt;
begin
    cc := 0;
    tt := 0;
    // get t8 thru t3
{$ifdef US}
    b := $00000001 and (t8c5 shr 9); tt := tt (b shl 8);
    b := $00000001 and (t8c5 shr 7); tt := tt + (b shl 7);
{$else} // Uk
    b := $00000001 and (t8c5 shr 9); tt := tt + (b shl 10);// US t8 -> UK t10 (Steve)
    b := $00000001 and (t8c5 shr 7); tt := tt + (b shl 8); // US t7 -> UK t8 (Doug & Steve)
{$endif}
    b := $00000001 and (t8c5 shr 5); tt := tt + (b shl 6);
    b := $00000001 and (t8c5 shr 4); tt := tt + (b shl 5);
    b := $00000001 and (t8c5 shr 3); tt := tt + (b shl 4);
    b := $00000001 and t8c5;         tt := tt + (b shl 3);
    // get c5 thru c2
{$ifdef US}
    b := $00000001 and (t8c5 shr 8); cc := cc + (b shl 5);
    b := $00000001 and (t8c5 shr 6); cc := cc + (b shl 4);
{$else} // Uk
    b := $00000001 and (t8c5 shr 8); tt := tt + (b shl 9); // US c5 -> UK t9 (Doug & Steve)
    b := $00000001 and (t8c5 shr 6); tt := tt + (b shl 7); // US c4 -> UK t7 (Doug)
{$endif}
    b := $00000001 and (t8c5 shr 2); cc := cc + (b shl 3);
    b := $00000001 and (t8c5 shr 1); cc := cc + (b shl 2) ;
    // get t2 thru t0
    b := $00000001 and (t2c1 shr 4); tt := tt + (b shl 2);
    b := $00000001 and (t2c1 shr 2); tt := tt + (b shl 1);
    b := $00000001 and t2c1;            tt := tt + b;
    // get c1 thru c0
    b := $00000001 and (t2c1 shr 3); cc := cc + (b shl 1);
    b := $00000001 and (t2c1 shr 1); cc := cc + b;

    tval := tt;
    cval := integer(cc);
end;

// ------------------------------------------------------------------- /
// Interleave : Used in Encoding                                       /
// ------------------------------------------------------------------- /
procedure Interleave(const tblidx: LongInt; const cval: integer; var t8c5: LongInt; var t2c1: integer);
var
   small,big,b: LongInt;
begin
	small := 0;
	big := 0;
	// build t2c1
	b := $00000001 and (tblidx shr 2); small := small + (b shl 4);
	b := $00000001 and (tblidx shr 1); small := small + (b shl 2);
	b := $00000001 and tblidx;	       small := small + b;
	b := $00000001 and (cval shr 1);   small := small + (b shl 3);
	b := $00000001 and cval;           small := small + (b shl 1);
	// build t8c5
{$ifdef US}
# ifdef US
	b := $00000001 and (tblidx shr 8);  big := big + (b shl 9);
	b := $00000001 and (tblidx shr 7);  big := big + (b shl 7);
{$else} // Uk
	b := $00000001 and (tblidx shr 10); big := big + (b shl 9);
	b := $00000001 and (tblidx shr 8);  big := big + (b shl 7);
{$endif}
	b := $00000001 and (tblidx shr 6);	big := big + (b shl 5);
	b := $00000001 and (tblidx shr 5);	big := big + (b shl 4);
	b := $00000001 and (tblidx shr 4);	big := big + (b shl 3);
	b := $00000001 and (tblidx shr 3);	big := big + b;
{$ifdef US}
	b := $00000001 and (cval shr 5);    big := big + (b shl 8);
	b := $00000001 and (cval shr 4) ;   big := big + (b shl 6);
{$else} // Uk
	b := $00000001 and (tblidx shr 9);  big := big + (b shl 8);
	b := $00000001 and (tblidx shr 7);  big := big + (b shl 6);
{$endif}

{$ifdef UKC4KNOWN}
	// experiment to find position for C4
	b := $00000001 and (cval shr 4);    big := big + (b shl 13);
{$endif}
	b := $00000001 and (cval shr 3);	big := big + (b shl 2);
	b := $00000001 and (cval shr 2);	big := big + (b shl 1);
{$ifdef UK}
	// for now, store the C4 and C5 bits way up high
    if (cval and $10) > 0 then big := big or C4_BIT;
    if (cval and $20) > 0 then big := big or C5_BIT;
{$endif}
	t8c5 := big;
	t2c1 := integer(small);
end;

// ------------------------------------------------------------------- /
// Lookup : Looksup a starttime and duration from global lookup table    /
//          at specified index. Used in decoding and encoding.           /
// ------------------------------------------------------------------- /
procedure Lookup(const i: LongInt; var outtime, outdur: integer);
begin
    if(initializer = 0) then begin
        initializer := 1;
        settables;
    end;
    if (i >= LIMIT) or (i < 0) then
       raise EXception.Create('Illegal lookup table index '+IntToStr(i));
    if (i >= SOFTLIMIT) then begin
        // anything above here is hypothetical anyway
        outtime := INVALID;
        outdur := INVALID;
        exit;
    end;
    outtime := slottable[i].start;
    outdur := slottable[i].leng;
end;

// ------------------------------------------------------------------- /
// SetTables : Loads lookup tables of start time and duration records.   /
//             Used in decoding and encoding.                            /
//                                                                     /
// Modified by PP : Populates first 192 from constant table            /
// ------------------------------------------------------------------- /
procedure SetTables;
var
    j: integer;
    dfltsptr: Ps_tab;
    sptr: Ps_tab;
begin
    // Populate first 192 entries from defaults table - PP
    dfltsptr := @dflt_slottable[0];
    sptr     := @slottable[0];
    for j := 0 to (NUM_DEFAULTS-1) do begin
        sptr.start := dfltsptr.start;
        sptr.leng  := dfltsptr.leng;
        inc(dfltsptr);
        inc(sptr);
    end;
    // Invalidate the remaining entries
    for j := NUM_DEFAULTS to (LIMIT-1) do begin
        sptr.start := INVALID;
        sptr.leng := INVALID;
        inc(sptr);
    end;

    fillQHB(192, 30);       // 30min progs starting 0015 -> 2345
    fillQHB(240, 60);
    fillQHB(288, 90);
    fillQHB(336, 120);
    fill(384, 2230, 30);    // 30 ->5 min progs starting 2230 -> 2255
    fill(418, 2300, 30);
    fill(452, 2330, 30);
    fill(486, 1930, 90);    // 90 -> 65min progs starting 1930 -> 1955
    fill(520, 2300, 90);    // 90 -> 65min progs starting 2300 -> 2325
    fill(554, 2330, 90);
    fill(588, 2130, 120);  // 120 -> 95min progs starting 2130 -> 2155
    fill(622, 2200, 120);
    fill(656, 2300, 120);  // 120 -> 95min progs starting 2300 -> 2325
    fill(690, 2330, 120);
    fill(724, 0, 120);
    fill(758, 30, 120);
    fill(792, 100, 120);
    fill(826, 130, 120);
    fill(860, 1730, 60);   // 60 -> 35min progs starting 1730 -> 1755
    fill(894, 1800, 60);
    fill(928, 1830, 60);
    fill(962, 1900, 60);
    fill(996, 1930, 60);
    fill(1030, 2000, 60);
    fill(1064, 2030, 60);
    fill(1098, 200, 120);  // 120 -> 95min progs starting 0200 -> 0225
    fill(1132, 230, 120);
    fill(1166, 300, 120);
    fill(1200, 330, 120);
    fill(1234, 400, 120);
    fill(1268, 1200, 120);
    fill(1302, 1400, 120);
    fill(1336, 1530, 120);
    fill(1370, 2100, 60);  // 60 -> 35min progs starting 2100 -> 2125
    fill(1404, 2130, 60);
    fill(1438, 2200, 60);
    fill(1472, 2230, 60);
    fill(1506, 2300, 60);
    fill(1540, 2330, 60);
    fill(1574, 1730, 30);  // 30 -> 5min progs starting 1730 -> 1755
    fill(1608, 1800, 30);
    fill(1642, 1830, 30);
    fill(1676, 1900, 30);
    fill(1710, 1930, 30);
    fill(1744, 2000, 30);
    fill(1778, 2030, 30);
    fill(1812, 2100, 30);
    fill(1846, 2130, 30);
    fill(1880, 2200, 30);
    fillHHB(1914, 150);    // 150min progs starting 2330 -> 0000
    fillHHB(1962, 180);    // 180min progs starting 2330 -> 0000
end;

// ------------------------------------------------------------------- /
// Fill : Populates 34 global slottable records, starting at specified /
//        index, with time and durations from global lists, adjusted by  /
//        passed amounts . Used in decoding and encoding.                /
// ------------------------------------------------------------------- /
procedure Fill(const index, time, dur: integer);
var
	i: integer;
    sptr: Ps_tab;
begin
    sptr := @slottable[index];
	for i := 0 to 33 do begin
		sptr.start := time + tlist[i]; // set adjusted time
		sptr.leng  := dur + dlist[i];  // set adjusted duration
		inc(sptr);                     // next record
	end;
	change := 1;
end;

// ------------------------------------------------------------------- /
// FillQHB : ???????                                                   /
//           Used in decoding and encoding.                              /
// ------------------------------------------------------------------- /
procedure FillQHB(const index, dur: integer);
var
	i: integer;
    sptr: Ps_tab;
begin
    sptr := @slottable[index];
    for i := 0 to 23 do begin
		sptr.start := i*100 + 15;
		sptr.leng := dur;
		inc(sptr);
		sptr.start := i*100 + 45;
		sptr.leng := dur;
		inc(sptr);
    end;
    change := 1;
end;


// ------------------------------------------------------------------- /
// FillHHB : ???????                                                   /
//           Used in decoding and encoding.                              /
// ------------------------------------------------------------------- /
procedure FillHHB(const index, dur: integer);
var
	i: integer;
    sptr: Ps_tab;
begin
    sptr := @slottable[index];
	for i := 23 downto 0 do begin
		sptr.start := i*100 + 30;
		sptr.leng := dur;
		inc(sptr);
		sptr.start := i*100;
		sptr.leng := dur;
		inc(sptr);
    end;
    change := 1;
end;


// ------------------------------------------------------------------- /
// EndTime : Returns end time based on start time and duration           /
//           Used in decoding.                                         /
// ------------------------------------------------------------------- /
function EndTime(start, dur: integer): integer;
begin
    Result := AddTime(start,dur) mod 2400;
end;

// ------------------------------------------------------------------- /
// FourDigitYear : Converts a date with a 2 digit year to one with a   /
//                 four digit year (eg. 99 -> 1999, 03 -> 2003)        /
// ------------------------------------------------------------------- /
procedure FourDigitYear(var ADate: TDate);
var
   Day, Month, Year: Word;
begin
   DecodeDate(ADate,Year,Month,Day);
   if Year > 1999 then exit;        // abort if already adjusted
   if Year = 99 then
      Year := 1999
   else
      Year := 2000 + Year;
   ADate := EncodeDate(Year,Month,Day);
end;

// ------------------------------------------------------------------- /
// BestTimes : ???????                                                 /
//             Used in encoding.                                       /
// ------------------------------------------------------------------- /
function BestTimes(var rstart, rduration: integer; var heuristic: Longword): LongInt;
var
	reqstart, reqend: integer;
	index, bestindex: LongInt;
	best, h, ds, de: Longword;
	startT, endT: integer;
begin
	reqstart := rstart;
	reqend := AddTime(reqstart,rduration);
	best := $FFFFFFFF;
    bestindex := best;
	for index := 0 to SOFTLIMIT-1 do begin
 	    Lookup(index,startT,endT);
		if startT > reqstart then continue;
		endT := AddTime(startT,endT);
		if endT < reqend then continue;
		// heuristic
		ds := diffminutes(reqstart,startT);
		de := diffminutes(endT,reqend);
		h := ds + de;
        if h < best then begin
			bestindex := index; // new best match
			best := h;
        end;
    end;
	// pass back values of rstart and rduration
	Lookup(bestindex,rstart,rduration);
	if (heuristic <> 0) then
		heuristic := best;
    Result := bestindex;
end;

// ------------------------------------------------------------------- /
// AddTime : ???????                                                   /
//           Used in decoding and encoding.                              /
// ------------------------------------------------------------------- /
function AddTime(const start, dur: integer): Integer;
var
	min, hr: integer;
begin
	min := (start mod 100) + (dur mod 60);
	hr := (start div 100) + (dur div 60);
	while min >= 60 do begin
		dec(min,60);
		inc(hr);
	end;
    Result := hr*100 + min;
end;

// ------------------------------------------------------------------- /
// DiffMinutes : ???????                                               /
//               Used in encoding .                                    /
// ------------------------------------------------------------------- /
function DiffMinutes(startT, endT: integer): integer;
begin
    startT := (startT div 100)*60 + (startT mod 100);
	endT := (endT div 100)*60 + (endT mod 100);
	Result := startT - endT;
end;

// ------------------------------------------------------------------- /
// ScanFor : ???????                                                   /
//           Used in encoding .                                        /
// ------------------------------------------------------------------- /
function ScanFor(const s, d: integer ): LongInt;
var
	j: LongInt;
begin
	if (initializer = 0) then begin
		initializer := 1;
		SetTables;
	end;
	for j := 0 to SOFTLIMIT-1 do
		if (slottable[j].start = s) and (slottable[j].leng = d) then begin
           Result := j; // Match found
           exit;
		end;
    Result := -1;       // nop match found
end;

// -------------------------------------------------------------------------------- /
// Decode : Converts a VideoPlus code (up to 6 digits only) to a programme schedule /
//                                                                                  /
// Input : VPCode  = VideoPlus code to convert (0 - 999999)                         /
//         RefDate = Reference date                                                 /
//         Schedule = structure to populate with results                            /
//                                                                                  /
// Output: Lookup table index                                                       /
// -------------------------------------------------------------------------------- /
function Decode(const VPCode: LongInt; const RefDate: TDate; var Sched: TSchedule): LongInt;
var
   Day, Month, Year: Word;
   s1_out, top5: Longint;
   bot3, quo, rem, s5_out, ofout: integer;
   mtout, tval: Longint;
   day_out, channel_out, starttime_out, endtime_out, duration_out: integer;
begin
   DecodeDate(RefDate,Year,Month,Day);
   Year := Year mod 100; // convert to 2 digit date
   if (VPCode < 1) or (VPCode > 999999) then
      raise Exception.Create('Invalid VideoPlus code (this program can only handle codes in the range 1 - 999999)');

   ofout := INVALID;
   mtout := INVALID;

   s1_out := f1(VPCode,KEY001);
   bot3 := integer(s1_out mod 1000);
   top5 := s1_out div 1000;
   quo := (bot3 - 1) div 32;
   rem := (bot3 - 1) mod 32;
   day_out := quo + 1;

	// spot instances of next month's/year's codes - S.Hosgood, 8.Feb.93
	if (day_out < day) then begin
		inc(month);
        if (month > 12) then begin
			month := 1;
			year := (year+1) mod 100;
		end;
	end;

	if VPCode >= 1000 then
		offset(day_out,year,top5,ofout,mtout)
	else begin
		mtout := 0;
		ofout := 0;
	end;
    s5_out := (rem + (day_out*(month+1)) + ofout) mod 32;
    DeInterleave(mtout, s5_out,tval,channel_out);
	inc(channel_out);
	Lookup(tval,starttime_out,duration_out);
    endtime_out := EndTime(starttime_out, duration_out);

    // Load schedule structure with decoded results
    Sched.SDate   := EncodeDate(year,month,day_out);
    FourDigitYear(Sched.SDate);
    Sched.STime   := EncodeTime(starttime_out div 100,starttime_out mod 100,0,0);
    Sched.ETime   := EncodeTime(endtime_out div 100,endtime_out mod 100,0,0);
    Sched.Durn    := duration_out;
    Sched.Channel := channel_out;
	Result := tval;
end;

// --------------------------------------------------------------------------------- /
// Encode : Converts a programme schedule into a VideoPlus code (up to 6 digits only)/
//                                                                                   /
// Input : RefDate = Reference date                                                  /
//         Schedule = structure with programme schedule                              /
//                                                                                   /
// Output: Returns VideoPlus code (0 - 999999)                                       /
// --------------------------------------------------------------------------------- /
function Encode(const RefDate: TDate; var Sched: TSchedule): LongInt;
var
    starttime: integer; // eg. 17:30 => 1730, 23:35 => 2335
	j: integer;
	tblidx, limit: LongInt;
	s5_out, bot3, rem, ofout: integer;
	month, day, year: word;
    hours, mins, secs, millis: word;
	s1_out, top5, s4_out: LongInt;
{$ifdef FULLSEARCH}
	long	doneflag;
{$else}
    i: integer;
    tmp: LongInt;
{$endif}
{$ifdef US}
	trailers: longint;
{$endif}
{$ifdef BEST}
	heuristic: Longword;
{$endif}
const
{$ifdef US}
    LATEST_START = 2330;
    MAX_CHANNEL  = 64;
{$else} // Uk
    LATEST_START = 2355;
    MAX_CHANNEL  = 16;
{$endif}
begin
    // Get start date and convert year to 2 digit year
    DecodeDate(Sched.SDate,year,month,day);
	year := year mod 100;
	if (month > 12) or (month < 1) or (day < 1) or (day > 31) then
      raise Exception.Create(FormatDateTime('Invalid start date ! c',Sched.SDate));
    // Convert start time and verify
    DecodeTime(Sched.STime,hours,mins, secs, millis);
	starttime := hours*100 + mins;
	if (starttime < 0) or (starttime > LATEST_START) then
      raise Exception.Create('Sorry, I cannot process that starting time !' +
                             #13#10 + FormatDateTime('c',Sched.STime));
    // Verify all other schedule parameters before encoding starts
	if (Sched.Channel < 1) or (Sched.Channel > MAX_CHANNEL) then
      raise Exception.Create('Invalid channel ! '+IntToStr(Sched.Channel));
{$ifdef US}
	trailers := starttime mod 100;
	if (trailers <> 0) and (trailers <> 30)) then // This will never be false !!! - PP
      raise Exception.Create('Invalid start time !' + #13#10 + FormatDateTime('c',Sched.STime));
	trailers := Sched.Durn mod 30;
	if ((trailers <> 0) or (Sched.Durn < 30) or (Sched.Durn > 300)) then
{$else} // Uk
	if (Sched.Durn < 0) or (Sched.Durn > 180) then
{$endif}
      raise Exception.Create('Sorry, I can''t process that program length ! '+IntToStr(Sched.Durn));

    // ============= Start the encoding ==================
	// get the t bits and the c bits
{$ifdef BEST}
	// do a best fit match on the timeslot
	tblidx := besttimes(starttime,Sched.Durn,heuristic);
	if heuristic <> 0 then
       raise Exception.Create(format('Exact match not available !'#10#13+
                                     'Best fit: %0.4d/%d',[starttime,Sched.Durn]));
{$else}
    tblidx := ScanFor(starttime,Sched.Durn);
	if tblidx < 0 then
       raise Exception.Create('Sorry, no index known');
{$endif}

	// from them infer what must have been step 4 and step 5 results
	Interleave(tblidx,Sched.Channel-1,s4_out,s5_out) ;
	// if the mapped_top is zero then top and offset are zero
	if s4_out = 0 then begin
		top5 := 0;
		ofout := 0;
	end else begin
{$ifdef FULLSEARCH}
	   // find the smallest unmapped_top giving correct mapped_top
       NumDigits(s4_out,limit);
	   top5 := limit div 10;
       doneflag := 0;
       while doneflag = 0 do begin
		   Offset(day,year,top5,&ofout,j);
           if j = s4_out then
              break;
           inc(top5);
   		   if top5 >= limit then
			  raise Exception.Create('Can''t seem to find an encoding - sorry');
		end; // while
{$else} // not FULLSEARCH
        j := NumDigits(s4_out,limit);
		if j > 3 then
			 raise Exception.Create('This schedule needs a 7 or 8 digit code - sorry');
		limit := limit div 10;
		ofout := 0;
		top5 := s4_out;
		// get a top5 with same no digits as s4_out, may have to loop several times
        repeat
			// reverse the map_top encryption
			top5 := UnmapTop(day,year,top5,j);
			// use that 'top5' to find offset
			for i := 0 to (year mod 16) do
				ofout := ofout + integer(MapTop(day,i,top5,j) mod 10);
		until not (top5 < limit);
		// add sum of final top5's digits to offset
		tmp := top5;
		while tmp > 0 do begin
			ofout := ofout + integer(tmp mod 10);
			tmp := tmp div 10;
        end;
		ofout := ofout mod 32;
{$endif}
	end;  // s4_out was not 0

	// have two of the three inputs to step 5; determine the rem
	for rem := 0 to 31 do begin
		j := (rem + (day*(month+1)) + ofout) mod 32;
		if j = s5_out then
           break;
    end;

	// assemble the output of step 1
	bot3 := EncodeBot3(day, rem);
	s1_out := bot3 + (1000 * top5) ;
	// invert the mixing and update result
	Result := F1(s1_out,KEY002);
end;

initialization
   initializer := 0;
   change      := 0;


end.
